#include <zephyr/types.h>
#include <stddef.h>
#include <string.h>
#include <errno.h>
#include <zephyr/sys/printk.h>
#include <zephyr/sys/byteorder.h>
#include <zephyr/kernel.h>

#include <zephyr/bluetooth/bluetooth.h>
#include <zephyr/bluetooth/hci.h>
#include <zephyr/bluetooth/conn.h>
#include <zephyr/bluetooth/uuid.h>
#include <zephyr/bluetooth/gatt.h>

#include <stdio.h>
#include <stdlib.h>

#include "../Vulcan_UUID/Vulcan_Uuid.h"
#include "../Vulcan_nvs/nvsStorage.h"
#include "../Vulcan_EMGSensor/EMGSensor.h"
#include "../Vulcan_RingState/RingState.h"

#define END_READ_SENSOR								"END_READ_SENSOR"
#define START_READ_SENSOR							"START_READ_SENSOR"

#define END_READ_FILTER_SENSOR				"END_READ_FILTER_SENSOR"
#define START_READ_FILTER_SENSOR			"START_READ_FILTER_SENSOR"

#define END_SEND_SENSOR								"END_SEND_SENSOR"
#define START_SEND_SENSOR							"START_SEND_SENSOR"

#define DATA_LEN 30

#define CHANELS     					3      					// samples

#define SIZE_ONE_PACKET_REV		sizeof(float) * (CHANELS + 1) // array CHANELS + 1 float (CHANELS sensor + 1 time read)
#define NUMBER_PACKET_REV			10							//10 packet/send
#define LENGTH_DATA_REV				NUMBER_PACKET_REV * SIZE_ONE_PACKET_REV // <600 byte

#define DATA_CALIB_LINE_MAX									50
#define SIZE_ONE_PACKET_CALIB_UINT16				5 	//THRESHOLD_SETUP | THRESHOLD_LOW | THRESHOLD_HIGH | TIME
#define SIZE_DATA_SAVE_CALIB_SD_UINT16			SIZE_ONE_PACKET_CALIB_UINT16 * DATA_CALIB_LINE_MAX			

#define SIZE_ONE_PACKET_CALIB_BYTE					SIZE_ONE_PACKET_CALIB_UINT16 * sizeof(uint16_t)
#define SIZE_DATA_SAVE_CALIB_SD_BYTE				SIZE_ONE_PACKET_CALIB_BYTE * DATA_CALIB_LINE_MAX

#define THRESHOLD_ID	4

uint8_t checkNotifyMain;
uint8_t signal_value[DATA_LEN + 1];
uint8_t threshold_value[DATA_LEN + 1];
uint8_t logic_value[DATA_LEN + 1];
uint8_t stateControl_value[DATA_LEN + 1];
uint8_t battery_value[DATA_LEN + 1];
uint8_t calibHistory_value[DATA_LEN + 1];

uint8_t onRead = 0;
uint8_t batteryPercent = 0;
uint16_t thresholdEMG[3] = {0, 0, 0};

void myoband_ccc_cfg_changed(const struct bt_gatt_attr *attr, uint16_t value)
{
	checkNotifyMain = value == BT_GATT_CCC_NOTIFY; 
}

ssize_t onReadSignal(struct bt_conn *conn, const struct bt_gatt_attr *attr,
			void *buf, uint16_t len, uint16_t offset)
{
	const char *value = attr->user_data;

	// int16_t count = emgSensor.readData(); //
	int16_t count = readSD();
	printk("Count readSD: %d\n", readSD());
	switch (count)
	{
		case -2:
			memcpy(value, START_SEND_SENSOR, strlen(START_SEND_SENSOR));
			break;
		case -1:
			memcpy(value, END_SEND_SENSOR, strlen(END_SEND_SENSOR));
			break;

		default:
			// pChar->setValue(emgSensor.rev, LENGTH_DATA_REV);
			memcpy(value, returnRev, LENGTH_DATA_REV);
			printk("Count: %d\n", count);
			break;
	}

	return bt_gatt_attr_read(conn, attr, buf, len, offset, value,
				 strlen(value));
}

ssize_t onWriteSignal(struct bt_conn *conn, const struct bt_gatt_attr *attr,
			 const void *buf, uint16_t len, uint16_t offset,
			 uint8_t flags)
{
	uint8_t *value = attr->user_data;
	char s[len+1];
	memset(s, 0, sizeof(s));
	memcpy(s, buf, len);
	printk("Write signal: %s\n", s);
	if (strlen(s) > 2) {
		char p[3];
		strncpy(p, s, 3);
		uint8_t timeRead = atoi(p);
		onRead = timeRead == 0 ? 0 : 1;
		setTimeReadSensor(timeRead);
	}

	if (offset + len > DATA_LEN) {
		return BT_GATT_ERR(BT_ATT_ERR_INVALID_OFFSET);
	}

	memcpy(value + offset, buf, len);
		// printk("%s\n", value);
	value[offset + len] = 0;

	return len;
}

ssize_t onReadThreshold(struct bt_conn *conn, const struct bt_gatt_attr *attr,
			void *buf, uint16_t len, uint16_t offset)
{
	const uint16_t *value = attr->user_data;
	memset((void *)thresholdEMG, 0, sizeof(thresholdEMG));
	memcpy((void *)thresholdEMG, getThreshold(), sizeof(thresholdEMG));
	memcpy((void*) value, (void*)thresholdEMG, sizeof(thresholdEMG));
	printk("Threshold: %d, %d, %d\n", thresholdEMG[0], thresholdEMG[1], thresholdEMG[2]);
	return bt_gatt_attr_read(conn, attr, buf, len, offset, value,
				 sizeof(thresholdEMG));
}

ssize_t onWriteThreshold(struct bt_conn *conn, const struct bt_gatt_attr *attr,
			 const void *buf, uint16_t len, uint16_t offset,
			 uint8_t flags)
{
	uint8_t *value = attr->user_data;

	char s[len+1];
	memset(s, 0, sizeof(s));
	memcpy(s, buf, len);
	if (strlen(s) > 11) {
		for (uint8_t i = 0; i < 3; i++) {
			char p[5];
			memcpy(p, s + i * 4, 4);
			p[4] = '\0';
			thresholdEMG[i] = atoi(p);
		}
		char pEpoch[10];
		strncpy(pEpoch, s + 3 * 4, 10);
		uint32_t epochTime = atoi(pEpoch);
		printk("Threshold: %d, %d, %d\n", thresholdEMG[0], thresholdEMG[1], thresholdEMG[2]);
		if (thresholdEMG[0] == 0) {
			printk("Calibration is false.\nYou must be calibration again!!!\n");
			notifyConfirm(false);
		}
		else {
			save_nvs_Threshold(THRESHOLD_ID, thresholdEMG, sizeof(thresholdEMG));
			// emgSensor.setThreshold(threshold, epochTime);
			notifyConfirm(true);
		}
	}

	if (offset + len > DATA_LEN) {
		return BT_GATT_ERR(BT_ATT_ERR_INVALID_OFFSET);
	}

	memcpy(value + offset, buf, len);
	value[offset + len] = 0;

	return len;
}

ssize_t onReadLogic(struct bt_conn *conn, const struct bt_gatt_attr *attr,
			void *buf, uint16_t len, uint16_t offset)
{
	const uint8_t *value = attr->user_data;
	uint8_t _modeLogic = getModeLogicControl();
	memcpy(value, &_modeLogic, sizeof(value));
	// printk("Value mode logic: %d\n", *value);

	return bt_gatt_attr_read(conn, attr, buf, len, offset, value,
				 strlen(value));
}

ssize_t onWriteLogic(struct bt_conn *conn, const struct bt_gatt_attr *attr,
			 const void *buf, uint16_t len, uint16_t offset,
			 uint8_t flags)
{
	uint8_t *value = attr->user_data;
	char s[len+1];
	memset(s, 0, sizeof(s));
	memcpy(s, buf, len);
	printk("Write logic: %s\n", s);
	if (strlen(s) > 0) {
		uint8_t modeLogic = atoi(s);
		// pref.putUShort("modeLogic", modeLogic);
		setModeLogicControl(modeLogic);
		notifyConfirm(true);
	}
	else {
		notifyConfirm(false);
	}

	if (offset + len > DATA_LEN) {
		return BT_GATT_ERR(BT_ATT_ERR_INVALID_OFFSET);
	}

	memcpy(value + offset, buf, len);
        // printk("%s\n", value);
	value[offset + len] = 0;

	return len;
}

ssize_t onReadStateControl(struct bt_conn *conn, const struct bt_gatt_attr *attr,
			void *buf, uint16_t len, uint16_t offset)
{
	const uint8_t *value = attr->user_data;
	uint8_t _stateControll = getStateControl() + 2;
	memcpy(value, _stateControll, sizeof(value));
	// printk("Read statecontrol: %d\n", *value);

	return bt_gatt_attr_read(conn, attr, buf, len, offset, value,
				 strlen(value));
}

ssize_t onReadBattery(struct bt_conn *conn, const struct bt_gatt_attr *attr,
			void *buf, uint16_t len, uint16_t offset)
{
	const uint8_t *value = attr->user_data;
	memcpy(value, &batteryPercent, sizeof(batteryPercent));
	// printk("Read Battery: %d\n", *value);

	return bt_gatt_attr_read(conn, attr, buf, len, offset, value,
				 sizeof(batteryPercent));
}

ssize_t onReadCalibHistory(struct bt_conn *conn, const struct bt_gatt_attr *attr,
			void *buf, uint16_t len, uint16_t offset)
{
	const char *value = attr->user_data;

	return bt_gatt_attr_read(conn, attr, buf, len, offset, value,
				 strlen(value));
}

ssize_t onWriteCalibHistory(struct bt_conn *conn, const struct bt_gatt_attr *attr,
			 const void *buf, uint16_t len, uint16_t offset,
			 uint8_t flags)
{
	uint8_t *value = attr->user_data;

	if (offset + len > DATA_LEN) {
		return BT_GATT_ERR(BT_ATT_ERR_INVALID_OFFSET);
	}

	memcpy(value + offset, buf, len);
	value[offset + len] = 0;

	return len;
}

BT_GATT_SERVICE_DEFINE(sensorService,
	BT_GATT_PRIMARY_SERVICE(&sensor_Service_UUID),
	BT_GATT_CHARACTERISTIC(&signal_UUID.uuid,
							BT_GATT_CHRC_READ | BT_GATT_CHRC_WRITE | BT_GATT_CHRC_NOTIFY,	
							BT_GATT_PERM_WRITE | BT_GATT_PERM_READ,	
							onReadSignal, onWriteSignal, signal_value),
	BT_GATT_CCC(myoband_ccc_cfg_changed, BT_GATT_PERM_READ | BT_GATT_PERM_WRITE),
	BT_GATT_CHARACTERISTIC(&threshold_UUID.uuid,
							BT_GATT_CHRC_READ | BT_GATT_CHRC_WRITE,
							BT_GATT_PERM_WRITE | BT_GATT_PERM_READ,	
							onReadThreshold, onWriteThreshold, threshold_value),
	BT_GATT_CHARACTERISTIC(&logic_UUID.uuid,
							BT_GATT_CHRC_READ | BT_GATT_CHRC_WRITE,
							BT_GATT_PERM_WRITE | BT_GATT_PERM_READ,	
							onReadLogic, onWriteLogic, logic_value),
	BT_GATT_CHARACTERISTIC(&state_Control_UUID.uuid,
							BT_GATT_CHRC_READ | BT_GATT_CHRC_NOTIFY,
							BT_GATT_PERM_WRITE | BT_GATT_PERM_READ,	
							onReadStateControl, NULL, stateControl_value),
	BT_GATT_CCC(myoband_ccc_cfg_changed, BT_GATT_PERM_READ | BT_GATT_PERM_WRITE),
	BT_GATT_CHARACTERISTIC(&battery_UUID.uuid,
							BT_GATT_CHRC_READ | BT_GATT_CHRC_NOTIFY,
							BT_GATT_PERM_WRITE | BT_GATT_PERM_READ,	
							onReadBattery, NULL, battery_value),
	BT_GATT_CCC(myoband_ccc_cfg_changed, BT_GATT_PERM_READ | BT_GATT_PERM_WRITE),
	BT_GATT_CHARACTERISTIC(&calib_History_UUID.uuid,
							BT_GATT_CHRC_READ | BT_GATT_CHRC_WRITE,
							BT_GATT_PERM_WRITE | BT_GATT_PERM_READ,	
							onReadCalibHistory, onWriteCalibHistory, calibHistory_value), 
);

void onNotifySignal(uint8_t _state) {
	// 1: firtRead
	// 2: onRead
	// 3: endRead
	switch (_state)
	{
		case 1:
			bt_gatt_notify_uuid(NULL, &signal_UUID.uuid, &sensorService.attrs[2], &START_READ_SENSOR, strlen(START_READ_SENSOR));
			break;
		case 3:
			bt_gatt_notify_uuid(NULL, &signal_UUID.uuid, &sensorService.attrs[2], &END_READ_SENSOR, strlen(END_READ_SENSOR));
			onRead = 0;
			break;
		case 2:
			bt_gatt_notify_uuid(NULL, &signal_UUID.uuid, &sensorService.attrs[2], returnData(), SIZE_ONE_PACKET_REV);
			break;
		default:
			break;
	}
}

void onNotifyStateControl(int8_t _stateControl) {
	uint8_t stateControl = _stateControl + 2;
	// printk("StateControl: %d\n", stateControl);
	bt_gatt_notify_uuid(NULL, &state_Control_UUID.uuid, &sensorService.attrs[2], &stateControl, 1);
}

void onNotifyBattery(uint8_t _battery) {
	if (batteryPercent == _battery) return;
	batteryPercent = _battery;
	bt_gatt_notify_uuid(NULL, &battery_UUID.uuid, &sensorService.attrs[2], &batteryPercent, sizeof(batteryPercent));
	// printk("Battery notify: %d\n", batteryPercent);
}

uint8_t getModeReadSensor() {
  	return onRead;
}

