#include <zephyr/kernel.h>
#include <zephyr/sys/reboot.h>
#include <zephyr/device.h>
#include <string.h>
#include <zephyr/drivers/flash.h>
#include <zephyr/storage/flash_map.h>
#include <zephyr/fs/nvs.h>

#define NVS_PARTITION		custom_nvs_storage
#define NVS_PARTITION_DEVICE	FIXED_PARTITION_DEVICE(NVS_PARTITION)
#define NVS_PARTITION_OFFSET	FIXED_PARTITION_OFFSET(NVS_PARTITION)

static struct nvs_fs *_fs;

void begin_NVS(struct nvs_fs* p_fs) {
    int rc = 0;
    _fs = p_fs;
	struct flash_pages_info info;	
	_fs->flash_device = NVS_PARTITION_DEVICE;
	if (!device_is_ready(_fs->flash_device)) {
		printk("Flash device %s is not ready\n", _fs->flash_device->name);
	}

	_fs->offset = NVS_PARTITION_OFFSET;
	rc = flash_get_page_info_by_offs(_fs->flash_device, _fs->offset, &info);
	if (rc) {
		printk("Unable to get page info\n");
	}
	_fs->sector_size = info.size;
	_fs->sector_count = 4;

	rc = nvs_mount(_fs);
	if (rc) {
		printk("Flash Init failed\n");
	}
}

void read_Init_DataString_Device(uint16_t _id, char *_dataInit, char *_dataNew, size_t len) {
	int rt = 0;
	rt = nvs_read(_fs, _id, _dataNew, len);
	if (rt > 0) {
		printk("Last ID: %d\t data: %s\n", _id, _dataNew);
	}else {
		nvs_write(_fs, _id, _dataInit, strlen(_dataInit));
		nvs_read(_fs, _id, _dataNew, len);
		printk("Update ID: %d\t data: %s \n", _id, _dataNew);
	}
}

void read_Init_DataString_Device_Version(uint16_t _id, char *_dataInit, char *_dataNow, size_t len) {
	int rt = 0;
	rt = nvs_read(_fs, _id, _dataNow, len);
	if (rt > 0) {
		if (strcmp(_dataNow, _dataInit) == 0) {
			printk("Last ID: %d\t data: %s\n", _id, _dataNow);
		}else {
			memcpy(_dataNow, _dataInit, strlen(_dataInit));
			nvs_write(_fs, _id, _dataInit, strlen(_dataInit));
			nvs_read(_fs, _id, _dataNow, len);
			printk("Update ID: %d\t data: %s \n", _id, _dataNow);
			NVIC_SystemReset();
		}
		
	}else {
		nvs_write(_fs, _id, _dataInit, strlen(_dataInit));
		nvs_read(_fs, _id, _dataNow, len);
		printk("Update ID: %d\t data: %s \n", _id, _dataNow);
	}
}

void read_Init_DataThreshold_Device(uint16_t _id, uint16_t *_dataInit, uint16_t *_dataNew, size_t len) {
	int rt = 0;
	uint16_t temp[2];
	rt = nvs_read(_fs, _id, _dataNew, len);
	if (rt > 0) {
		memcpy(temp, _dataNew, len);
		printk("Last ID: %d\t Threshold CH1 %d   CH2 %d\n", _id, temp[0], temp[1]);
	}else {
		(void)nvs_write(_fs, _id, _dataInit, len);
		nvs_read(_fs, _id, _dataNew, len);
		memcpy(temp, _dataNew, len);
		printk("Update ID: %d\t Threshold CH1 %d   CH2 %d\n", _id, temp[0], temp[1]);
	}
}

void read_Init_DataInt_Device(uint16_t _id, uint8_t *_dataInit, uint8_t *_dataNew, size_t len) {
	int rt = 0;
	rt = nvs_read(_fs, _id, _dataNew, sizeof(_dataNew));
	if (rt > 0) {
		printk("Last ID: %d\t Data: %d\n", _id, *_dataNew);
	}else {
		(void)nvs_write(_fs, _id, _dataInit, len);
		nvs_read(_fs, _id, _dataNew, len);
		printk("Update ID: %d\t Data: %d\n", _id, *_dataNew);
	}
}

void save_nvs_String(uint16_t id, char *pData) {
	printk("Data Save %s\n", pData);
	nvs_write(_fs, id, pData, strlen(pData));
	nvs_read(_fs, id, pData, sizeof(pData));
	printk("Update data:  %s\n", pData);
}

void save_nvs_Int16(uint16_t id, uint16_t *pData) {
	printk("Data Save %d\n", *pData);
	nvs_write(_fs, id, pData, sizeof(pData));
	nvs_read(_fs, id, pData, sizeof(pData));
	printk("Update data:  %d\n", *pData);
}

void save_nvs_Int8(uint16_t id, uint8_t *pData) {
	printk("Data Save %d\n", *pData);
	nvs_write(_fs, id, pData, sizeof(pData));
	nvs_read(_fs, id, pData, sizeof(pData));
	printk("Update data:  %d\n", *pData);
}

uint8_t get_nvs_Threshold(uint16_t id, uint16_t *pdata, size_t len) {
	uint8_t rt = (nvs_read(_fs, id, pdata, len) > 0)? 1:0;
	uint16_t temp[3];
	if (rt > 0) {
		memcpy(temp, pdata, len);
		printk("Last ID Threshold: %d\t Data: %d, %d, %d\n", id, temp[0], temp[1], temp[2]);
	}
	return rt;
}

void save_nvs_Threshold(uint16_t id, uint16_t *pData, uint8_t lenData) {
	printk("Data Save %d\n", *pData);
	nvs_write(_fs, id, pData, lenData);
	nvs_read(_fs, id, pData, lenData);
	printk("Update data:  %d\n", *pData);
}

uint8_t get_nvs_ModeLogic(uint16_t id, uint8_t *pdata, size_t len) {
	int rt = nvs_read(_fs, id, pdata, len);
	uint16_t temp[2];
	if (rt > 0) {
		memcpy(temp, pdata, len);
		printk("Mode logic: %d\n", temp[0]);
		return temp[0];
	}
	return 1;
}